import { Injectable } from '@angular/core';
import { BehaviorSubject, Observable, first, map, startWith, tap } from 'rxjs';
import { Organisationseinheit, OrganisationseinheitState } from './user.model';
import { UserRepository } from './user.repository.service';

@Injectable({
  providedIn: 'root',
})
export class UserService {
  readonly organisationseinheitState$: BehaviorSubject<OrganisationseinheitState> =
    new BehaviorSubject({
      organisationseinheitItems: [],
      loading: false,
      updateRequired: true,
    });

  constructor(private userRepository: UserRepository) {}

  public saveOrganisationseinheit(organisationseinheit: Organisationseinheit): Observable<boolean> {
    return this.setLoadingUntilFirst(
      this.userRepository.saveOrganisationseinheit(organisationseinheit),
      () => {
        this.updateExistingOrganisationseinheit(organisationseinheit);
      },
    );
  }

  private updateExistingOrganisationseinheit(organisationseinheit: Organisationseinheit): void {
    this.setOrganisationseinheit([
      organisationseinheit,
      ...this.organisationseinheitState$.value.organisationseinheitItems.filter(
        (otherOrganisationseinheit: Organisationseinheit) =>
          otherOrganisationseinheit.id !== organisationseinheit.id,
      ),
    ]);
  }

  public createOrganisationseinheit(
    name: string,
    organisationseinheitIds: string[],
  ): Observable<boolean> {
    return this.setLoadingUntilFirst(
      this.userRepository.createOrganisationseinheit(name, organisationseinheitIds),
      (organisationseinheit: Organisationseinheit) => {
        this.addOrganisationseinheit(organisationseinheit);
      },
    );
  }

  public deleteOrganisationseinheit(id: string): Observable<boolean> {
    return this.setLoadingUntilFirst(this.userRepository.deleteOrganisationseinheit(id), () =>
      this.removeOrganisationseinheit(id),
    );
  }

  setLoadingUntilFirst<T>(
    action: Observable<T>,
    tapFunction: (value: T) => void,
  ): Observable<boolean> {
    this.setLoading();

    return action.pipe(
      first(),
      tap(tapFunction),
      tap(() => this.setLoading(false)),
      map(() => this.organisationseinheitState$.value.loading),
      startWith(this.organisationseinheitState$.value.loading),
    );
  }

  private removeOrganisationseinheit(id: string): void {
    this.setOrganisationseinheit(
      this.organisationseinheitState$.value.organisationseinheitItems.filter(
        (organisationseinheit: Organisationseinheit) => organisationseinheit.id !== id,
      ),
    );
  }

  private addOrganisationseinheit(organisationseinheit: Organisationseinheit): void {
    this.setOrganisationseinheit([
      ...this.organisationseinheitState$.value.organisationseinheitItems,
      organisationseinheit,
    ]);
  }

  public getOrganisationseinheitItems(): Observable<Organisationseinheit[]> {
    return this.getOrganisationseinheitState().pipe(
      map((state: OrganisationseinheitState) => state.organisationseinheitItems),
    );
  }

  public getOrganisationseinheitState(): Observable<OrganisationseinheitState> {
    return this.organisationseinheitState$.asObservable().pipe(tap(() => this.updateIfRequired()));
  }

  updateIfRequired(): void {
    if (this.isUpdateRequired(this.organisationseinheitState$.value)) {
      this.updateOrganisationseinheitItems().pipe(first()).subscribe();
    }
  }

  private isUpdateRequired(state: OrganisationseinheitState): boolean {
    return state.updateRequired && !state.loading;
  }

  updateOrganisationseinheitItems(): Observable<Organisationseinheit[]> {
    this.setLoading();

    return this.userRepository.findOrganisationseinheitItems().pipe(
      tap(() => this.setUpdateRequired(false)),
      tap((organisationseinheitItems: Organisationseinheit[]) =>
        this.setOrganisationseinheit(organisationseinheitItems),
      ),
    );
  }

  setLoading(loading: boolean = true): void {
    this.organisationseinheitState$.next({
      ...this.organisationseinheitState$.value,
      loading,
    });
  }

  setUpdateRequired(updateRequired: boolean = true) {
    this.organisationseinheitState$.next({
      ...this.organisationseinheitState$.value,
      updateRequired,
    });
  }

  setOrganisationseinheit(organisationseinheitItems: Organisationseinheit[]): void {
    this.organisationseinheitState$.next({
      ...this.organisationseinheitState$.value,
      organisationseinheitItems: this.sortedOrganisationseinheitItems(organisationseinheitItems),
    });
  }

  private sortedOrganisationseinheitItems(
    organisationseinheitItems: Organisationseinheit[],
  ): Organisationseinheit[] {
    return [...organisationseinheitItems].sort((a, b) => a.name.localeCompare(b.name));
  }
}
