import {
  BinaryFileListResource,
  BinaryFileResource,
  BinaryFileService,
} from '@alfa-client/binary-file-shared';
import {
  CommandOrder,
  CommandResource,
  CommandService,
  CreateCommandProps,
} from '@alfa-client/command-shared';
import {
  ApiError,
  EMPTY_STRING,
  HttpError,
  StateResource,
  createEmptyStateResource,
  createErrorStateResource,
  createStateResource,
} from '@alfa-client/tech-shared';
import { Mock, mock, useFromMock } from '@alfa-client/test-utils';
import {
  VorgangCommandService,
  VorgangService,
  VorgangWithEingangLinkRel,
  VorgangWithEingangResource,
} from '@alfa-client/vorgang-shared';
import { fakeAsync, tick } from '@angular/core/testing';
import faker from '@faker-js/faker';
import { ResourceUri, getUrl } from '@ngxp/rest';
import { cold } from 'jest-marbles';
import { CommandLinkRel } from 'libs/command-shared/src/lib/command.linkrel';
import { createApiError } from 'libs/tech-shared/test/error';
import { createVorgangWithEingangResource } from 'libs/vorgang-shared/test/vorgang';
import { Observable, first, of } from 'rxjs';
import {
  createBinaryFileListResource,
  createBinaryFileResource,
} from '../../../binary-file-shared/test/binary-file';
import {
  createCommandResource,
  createCommandStateResource,
  createCreateCommandProps,
} from '../../../command-shared/test/command';
import { ResourceRepository } from '../../../tech-shared/src/lib/resource/resource.repository';
import { createFile } from '../../../tech-shared/test/file';
import { singleCold, singleColdCompleted } from '../../../tech-shared/test/marbles';
import {
  createBescheid,
  createBescheidListResource,
  createBescheidResource,
  createBescheidStateResource,
  createUploadFileInProgress,
} from '../test/bescheid';
import { createDocumentResource } from '../test/document';
import { BescheidFacade } from './+state/bescheid.facade';
import { BescheidLinkRel } from './bescheid.linkrel';
import {
  Bescheid,
  BescheidListResource,
  BescheidResource,
  BescheidStatus,
  UploadFileInProgress,
} from './bescheid.model';
import { BescheidService } from './bescheid.service';
import { DocumentLinkRel } from './document.linkrel';
import { DocumentResource } from './document.model';

import * as DateUtil from '../../../tech-shared/src/lib/date.util';
import * as BescheidUtil from './bescheid.util';

describe('BescheidService', () => {
  let service: BescheidService;

  let facade: Mock<BescheidFacade>;
  let vorgangService: Mock<VorgangService>;
  let resourceRepository: Mock<ResourceRepository>;
  let commandService: Mock<CommandService>;
  let vorgangCommandService: Mock<VorgangCommandService>;
  let binaryFileService: Mock<BinaryFileService>;

  const vorgangWithEingangStateResource: StateResource<VorgangWithEingangResource> =
    createStateResource(createVorgangWithEingangResource());

  beforeEach(() => {
    facade = mock(BescheidFacade);
    resourceRepository = mock(ResourceRepository);
    commandService = mock(CommandService);
    vorgangCommandService = mock(VorgangCommandService);

    vorgangService = mock(VorgangService);
    vorgangService.getVorgangWithEingang.mockReturnValue(of(vorgangWithEingangStateResource));

    binaryFileService = mock(BinaryFileService);

    service = new BescheidService(
      useFromMock(facade),
      useFromMock(vorgangService),
      useFromMock(commandService),
      useFromMock(vorgangCommandService),
      useFromMock(binaryFileService),
      useFromMock(resourceRepository),
    );
  });

  it('should be created', () => {
    expect(service).toBeTruthy();
  });

  describe('get bescheid draft', () => {
    const bescheidDraft: BescheidResource = createBescheidResource();
    const bescheidDraftStateResource: StateResource<BescheidResource> =
      createStateResource(bescheidDraft);

    it('should call resource service', () => {
      service.bescheidDraftService.get = jest.fn();

      service.getBescheidDraft();

      expect(service.bescheidDraftService.get).toHaveBeenCalled();
    });

    it('should return value', () => {
      service.bescheidDraftService.get = jest
        .fn()
        .mockReturnValue(singleCold(bescheidDraftStateResource));

      const bescheidStateResource$: Observable<StateResource<BescheidResource>> =
        service.getBescheidDraft();

      expect(bescheidStateResource$).toBeObservable(singleCold(bescheidDraftStateResource));
    });
  });

  describe('getBescheidCommand', () => {
    it('should call facade', () => {
      service.getBescheidCommand();

      expect(facade.getBescheidCommand).toHaveBeenCalled();
    });
  });

  describe('createBescheid', () => {
    const vorgangWithEingang: VorgangWithEingangResource = createVorgangWithEingangResource();
    const command: CommandResource = createCommandResource([CommandLinkRel.EFFECTED_RESOURCE]);
    const commandStateResource: StateResource<CommandResource> = createStateResource(command);

    beforeEach(() => {
      facade.getBescheidCommand.mockReturnValue(of(commandStateResource));
      service.bescheidDraftService.setResourceByUri = jest.fn();
    });

    it('should call facade', () => {
      service.createBescheid(vorgangWithEingang).pipe(first()).subscribe();

      expect(facade.createBescheid).toHaveBeenCalledWith(vorgangWithEingang, {
        order: CommandOrder.CREATE_BESCHEID,
        body: null,
      });
    });

    it('should set resource by uri', () => {
      service.createBescheid(vorgangWithEingang).pipe(first()).subscribe();

      expect(service.bescheidDraftService.setResourceByUri).toHaveBeenCalledWith(
        getUrl(command, CommandLinkRel.EFFECTED_RESOURCE),
      );
    });
  });

  describe('bescheidErstellungUeberspringen', () => {
    describe('Bescheid Draft exists', () => {
      const vorgangWithEingangResource: VorgangWithEingangResource =
        createVorgangWithEingangResource();

      const bescheidResource: BescheidResource = createBescheidResource();
      const bescheidStateResource: StateResource<BescheidResource> =
        createStateResource(bescheidResource);

      const commandStateResource: StateResource<CommandResource> = createCommandStateResource();

      beforeEach(() => {
        service.existsBescheidDraft = jest.fn().mockReturnValue(true);
        service.getBescheidDraft = jest.fn().mockReturnValue(of(bescheidStateResource));
        service.bescheidLoeschenUndErstellungUeberspringen = jest
          .fn()
          .mockReturnValue(of(commandStateResource));
      });

      it('should get bescheid draft', (done) => {
        const command$: Observable<StateResource<CommandResource>> =
          service.bescheidErstellungUeberspringen(vorgangWithEingangResource);

        command$.subscribe(() => {
          expect(service.getBescheidDraft).toHaveBeenCalled();
          done();
        });
      });

      it('should Bescheid löschen und Erstellung überspringen', (done) => {
        const command$: Observable<StateResource<CommandResource>> =
          service.bescheidErstellungUeberspringen(vorgangWithEingangResource);

        command$.subscribe(() => {
          expect(service.bescheidLoeschenUndErstellungUeberspringen).toHaveBeenCalledWith(
            vorgangWithEingangResource,
            bescheidResource,
          );
          done();
        });
      });

      it('should return command', () => {
        const command$: Observable<StateResource<CommandResource>> =
          service.bescheidErstellungUeberspringen(vorgangWithEingangResource);

        expect(command$).toBeObservable(cold('(a|)', { a: commandStateResource }));
      });
    });

    describe('Bescheid Draft not exists', () => {
      const vorgangWithEingangResource: VorgangWithEingangResource =
        createVorgangWithEingangResource();
      const commandStateResource: StateResource<CommandResource> = createCommandStateResource();

      beforeEach(() => {
        service.existsBescheidDraft = jest.fn().mockReturnValue(false);
        service.vorgangAbschliesen = jest.fn().mockReturnValue(of(commandStateResource));
      });

      it('should call vorgang abschliessen', (done) => {
        const command$: Observable<StateResource<CommandResource>> =
          service.bescheidErstellungUeberspringen(vorgangWithEingangResource);

        command$.subscribe(() => {
          expect(service.vorgangAbschliesen).toHaveBeenCalledWith(vorgangWithEingangResource);
          done();
        });
      });

      it('should return command', () => {
        const command$: Observable<StateResource<CommandResource>> =
          service.bescheidErstellungUeberspringen(vorgangWithEingangResource);

        expect(command$).toBeObservable(cold('(a|)', { a: commandStateResource }));
      });
    });
  });

  describe('getBescheidDraftIfExists', () => {
    let vorgangStateResource: StateResource<VorgangWithEingangResource>;
    let bescheidStateResource: StateResource<BescheidResource>;

    beforeEach(() => {
      vorgangStateResource = createStateResource(
        createVorgangWithEingangResource([VorgangWithEingangLinkRel.BESCHEID_DRAFT]),
      );
      bescheidStateResource = createBescheidStateResource();
      vorgangService.getVorgangWithEingang.mockReturnValue(of(vorgangStateResource));
      service.getBescheidDraft = jest.fn().mockReturnValue(of(bescheidStateResource));
    });

    it('should get vorgang with eingang', () => {
      service.getBescheidDraftIfExists();

      expect(vorgangService.getVorgangWithEingang).toHaveBeenCalled();
    });

    it('should emit state resources', () => {
      const bescheid$: Observable<StateResource<BescheidResource>> =
        service.getBescheidDraftIfExists();

      expect(bescheid$).toBeObservable(
        cold('(ab|)', { a: createEmptyStateResource(), b: bescheidStateResource }),
      );
    });
  });

  describe('bescheidLoeschenUndErstellungUeberspringen', () => {
    const vorgangWithEingangResource: VorgangWithEingangResource =
      createVorgangWithEingangResource();
    const bescheidResource: BescheidResource = createBescheidResource();
    const vorgangAbschliessenCommandStateResource: StateResource<CommandResource> =
      createCommandStateResource();

    beforeEach(() => {
      service.vorgangAbschliesen = jest
        .fn()
        .mockReturnValue(of(vorgangAbschliessenCommandStateResource));
      service.deleteBescheid = jest.fn().mockReturnValue(of(createCommandStateResource));
    });

    it('should Bescheiderstellung überspringen', (done) => {
      const command$: Observable<StateResource<CommandResource>> =
        service.bescheidLoeschenUndErstellungUeberspringen(
          vorgangWithEingangResource,
          bescheidResource,
        );

      command$.subscribe(() => {
        expect(service.vorgangAbschliesen).toHaveBeenCalledWith(vorgangWithEingangResource);
        done();
      });
    });

    it('should Bescheid löschen', (done) => {
      service.vorgangAbschliesen = jest
        .fn()
        .mockReturnValue(of(createCommandStateResource([CommandLinkRel.EFFECTED_RESOURCE])));

      const command$: Observable<StateResource<CommandResource>> =
        service.bescheidLoeschenUndErstellungUeberspringen(
          vorgangWithEingangResource,
          bescheidResource,
        );

      command$.subscribe(() => {
        expect(service.deleteBescheid).toHaveBeenCalledWith(bescheidResource);
        done();
      });
    });

    it('should not Bescheid löschen', (done) => {
      service.vorgangAbschliesen = jest.fn().mockReturnValue(of(createCommandStateResource()));

      const command$: Observable<StateResource<CommandResource>> =
        service.bescheidLoeschenUndErstellungUeberspringen(
          vorgangWithEingangResource,
          bescheidResource,
        );

      command$.subscribe(() => {
        expect(service.deleteBescheid).not.toHaveBeenCalledWith(bescheidResource);
        done();
      });
    });

    it('should emit vorgang abschliessen command', () => {
      const command$: Observable<StateResource<CommandResource>> =
        service.bescheidLoeschenUndErstellungUeberspringen(
          vorgangWithEingangResource,
          bescheidResource,
        );

      expect(command$).toBeObservable(cold('(a|)', { a: vorgangAbschliessenCommandStateResource }));
    });
  });

  describe('vorgang abschliessen', () => {
    const vorgangWithEingangResource: VorgangWithEingangResource =
      createVorgangWithEingangResource();
    const commandStateResource: StateResource<CommandResource> = createCommandStateResource();

    beforeEach(() => {
      vorgangCommandService.abschliessen.mockReturnValue(of(commandStateResource));
    });

    it('should call vorgang command service', (done) => {
      const command$: Observable<StateResource<CommandResource>> = service.vorgangAbschliesen(
        vorgangWithEingangResource,
      );

      command$.subscribe(() => {
        expect(vorgangCommandService.abschliessen).toHaveBeenCalledWith(vorgangWithEingangResource);
        done();
      });
    });

    it('should return command', () => {
      const command$: Observable<StateResource<CommandResource>> =
        service.bescheidErstellungUeberspringen(vorgangWithEingangResource);

      expect(command$).toBeObservable(cold('(a|)', { a: commandStateResource }));
    });
  });

  describe('delete bescheid', () => {
    const bescheidResource: BescheidResource = createBescheidResource();

    it('should create command', () => {
      service.deleteBescheid(bescheidResource);

      const expectedProps: CreateCommandProps = {
        resource: bescheidResource,
        linkRel: BescheidLinkRel.DELETE,
        command: {
          order: CommandOrder.DELETE_BESCHEID,
          body: null,
        },
        snackBarMessage: EMPTY_STRING,
      };
      expect(commandService.createCommandByProps).toHaveBeenCalledWith(expectedProps);
    });

    it('should return command', () => {
      const commandStateResource: StateResource<CommandResource> = createEmptyStateResource();
      commandService.createCommandByProps.mockReturnValue(commandStateResource);

      const createdCommand: Observable<StateResource<CommandResource>> =
        service.deleteBescheid(bescheidResource);

      expect(createdCommand).toEqual(commandStateResource);
    });
  });

  describe('update bescheid', () => {
    const bescheidResource: BescheidResource = createBescheidResource();
    const bescheid: Bescheid = createBescheid();
    const commandResource: CommandResource = createCommandResource([
      CommandLinkRel.EFFECTED_RESOURCE,
    ]);
    const commandStateResource: StateResource<CommandResource> =
      createStateResource(commandResource);
    const createCommandProps: CreateCommandProps = createCreateCommandProps();
    let buildUpdateBescheidCommandPropsSpy: jest.SpyInstance;

    beforeEach(() => {
      buildUpdateBescheidCommandPropsSpy = jest
        .spyOn(BescheidUtil, 'buildUpdateBescheidCommandProps')
        .mockReturnValue(createCommandProps);
      service.bescheidDraftService.stateResource.next(createStateResource(bescheidResource));
      commandService.createCommandByProps.mockReturnValue(of(commandStateResource));
      service.bescheidDraftService.setResourceByUri = jest.fn();
    });

    it('should build update bescheid command props', () => {
      service.updateBescheid(bescheid);

      expect(buildUpdateBescheidCommandPropsSpy).toHaveBeenCalledWith(
        service.bescheidDraftService.getResource(),
        bescheid,
      );
    });

    it('should create command', () => {
      service.updateBescheid(bescheid);

      expect(commandService.createCommandByProps).toHaveBeenCalledWith(createCommandProps);
    });

    it('should return command', () => {
      const updateBescheid$: Observable<StateResource<CommandResource>> =
        service.updateBescheid(bescheid);

      expect(updateBescheid$).toBeObservable(cold('(a|)', { a: commandStateResource }));
    });

    it('should set resource by uri', (done) => {
      service
        .updateBescheid(bescheid)
        .pipe(first())
        .subscribe((commandStateResource: StateResource<CommandResource>) => {
          expect(service.bescheidDraftService.setResourceByUri).toHaveBeenCalledWith(
            getUrl(commandStateResource.resource, CommandLinkRel.EFFECTED_RESOURCE),
          );
          done();
        });
    });

    it('should clear create bescheid document in progress', (done) => {
      service.createBescheidDocumentInProgress$.next(createUploadFileInProgress());

      service.updateBescheid(bescheid).subscribe(() => {
        expect(service.createBescheidDocumentInProgress$.value).toEqual({ loading: false });
        done();
      });
    });

    it('should clear upload bescheid document in progress', (done) => {
      service.uploadBescheidDocumentInProgress$.next(createUploadFileInProgress());

      service.updateBescheid(bescheid).subscribe(() => {
        expect(service.uploadBescheidDocumentInProgress$.value).toEqual({ loading: false });
        done();
      });
    });
  });

  describe('send bescheid', () => {
    const bescheidResource: BescheidResource = createBescheidResource();
    const createCommandProps: CreateCommandProps = createCreateCommandProps();
    const commandStateResource: StateResource<CommandResource> =
      createStateResource(createCommandResource());
    const linkRel: string = 'link_rel';

    let buildSendBescheidCommandPropsSpy: jest.SpyInstance;

    beforeEach(() => {
      service.bescheidDraftService.get = jest
        .fn()
        .mockReturnValue(of(createStateResource(bescheidResource)));
      buildSendBescheidCommandPropsSpy = jest
        .spyOn(BescheidUtil, 'buildSendBescheidCommandProps')
        .mockReturnValue(createCommandProps);
      commandService.createCommandByProps.mockReturnValue(of(commandStateResource));
    });

    it('should get resource', () => {
      service.sendBescheid(bescheidResource, linkRel);

      expect(service.bescheidDraftService.get).toHaveBeenCalled();
    });

    it('should call command service', (done) => {
      service.sendBescheid(bescheidResource, linkRel).subscribe(() => {
        expect(commandService.createCommandByProps).toHaveBeenCalledWith(createCommandProps);
        done();
      });
    });

    it('should build send bescheid command props', (done) => {
      service.sendBescheid(bescheidResource, linkRel).subscribe(() => {
        expect(buildSendBescheidCommandPropsSpy).toHaveBeenCalledWith(bescheidResource, linkRel);
        done();
      });
    });

    it('should return command', () => {
      const command$: Observable<StateResource<CommandResource>> = service.sendBescheid(
        bescheidResource,
        linkRel,
      );

      expect(command$).toBeObservable(cold('(a|)', { a: commandStateResource }));
    });
  });

  describe('sendBescheidManually', () => {
    const bescheidResource: BescheidResource = createBescheidResource();
    const sendBescheidCommand: StateResource<CommandResource> =
      createStateResource(createCommandResource());

    beforeEach(() => {
      service.sendBescheid = jest.fn().mockReturnValue(of(sendBescheidCommand));
    });

    it('should call sendBescheid', () => {
      service.sendBescheidManually(bescheidResource);

      expect(service.sendBescheid).toHaveBeenCalledWith(
        bescheidResource,
        BescheidLinkRel.BESCHEIDEN,
      );
    });

    it('should return command', () => {
      const command$: Observable<StateResource<CommandResource>> =
        service.sendBescheidManually(bescheidResource);

      expect(command$).toBeObservable(singleColdCompleted(sendBescheidCommand));
    });
  });

  describe('sendBescheidToAntragsteller', () => {
    const bescheidResource: BescheidResource = createBescheidResource();
    const sendBescheidCommand: StateResource<CommandResource> =
      createStateResource(createCommandResource());

    beforeEach(() => {
      service.sendBescheid = jest.fn().mockReturnValue(of(sendBescheidCommand));
    });

    it('should call sendBescheid', () => {
      service.sendBescheidToAntragsteller(bescheidResource);

      expect(service.sendBescheid).toHaveBeenCalledWith(
        bescheidResource,
        BescheidLinkRel.BESCHEIDEN_UND_SENDEN,
      );
    });

    it('should return command', () => {
      const command$: Observable<StateResource<CommandResource>> =
        service.sendBescheidToAntragsteller(bescheidResource);

      expect(command$).toBeObservable(singleColdCompleted(sendBescheidCommand));
    });
  });

  describe('do update bescheid', () => {
    const bescheid: Bescheid = createBescheid();
    const bescheidResource: BescheidResource = createBescheidResource();

    const createCommandProps: CreateCommandProps = createCreateCommandProps();
    let buildUpdateBescheidCommandPropsSpy: jest.SpyInstance;

    beforeEach(() => {
      buildUpdateBescheidCommandPropsSpy = jest
        .spyOn(BescheidUtil, 'buildUpdateBescheidCommandProps')
        .mockReturnValue(createCommandProps);
      commandService.createCommandByProps.mockClear();
      commandService.createCommandByProps.mockReturnValue(
        of(createStateResource(createCommandResource())),
      );
    });

    it('should build update bescheid command props', () => {
      service.doUpdateBescheid(bescheidResource, bescheid);

      expect(buildUpdateBescheidCommandPropsSpy).toHaveBeenCalledWith(bescheidResource, bescheid);
    });

    it('should call command service', () => {
      service.doUpdateBescheid(bescheidResource, bescheid).subscribe();

      expect(commandService.createCommandByProps).toHaveBeenCalledWith(createCommandProps);
    });
  });

  describe('getAttachments', () => {
    let bescheidResource: BescheidResource;
    let binaryFileListResource: BinaryFileListResource;
    let binaryFileResource: BinaryFileResource;

    beforeEach(() => {
      bescheidResource = createBescheidResource([BescheidLinkRel.ATTACHMENTS]);
      service.getBescheidDraft = jest
        .fn()
        .mockReturnValue(of(createStateResource(bescheidResource)));

      binaryFileResource = createBinaryFileResource();

      binaryFileListResource = createBinaryFileListResource([binaryFileResource]);
      binaryFileService.getFiles.mockReturnValue(of(createStateResource(binaryFileListResource)));
    });

    it('should get files', (done) => {
      service.getAttachments().subscribe(() => {
        expect(binaryFileService.getFiles).toHaveBeenCalledWith(
          bescheidResource,
          BescheidLinkRel.ATTACHMENTS,
        );
        done();
      });
    });

    it('should return embedded resources', () => {
      const attachments$: Observable<BinaryFileResource[]> = service.getAttachments();

      expect(attachments$).toBeObservable(singleColdCompleted([binaryFileResource]));
    });
  });

  describe('get document uri', () => {
    const documentUri: ResourceUri = faker.internet.url();

    beforeEach(() => {
      service.bescheidDocumentUri$.next(documentUri);
    });

    it('should return value', (done) => {
      service.getDocumentUri().subscribe((uri) => {
        expect(uri).toBe(documentUri);
        done();
      });
    });
  });

  describe('get bescheid document file', () => {
    it('should return value', (done) => {
      const binaryFile: BinaryFileResource = createBinaryFileResource();
      const binaryFileStateResource: StateResource<BinaryFileResource> =
        createStateResource(binaryFile);
      service.bescheidDocumentFile$.next(binaryFileStateResource);

      service.getBescheidDocumentFile().subscribe((result: StateResource<BinaryFileResource>) => {
        expect(result).toBe(binaryFileStateResource);
        done();
      });
    });
  });

  describe('upload bescheid document', () => {
    const bescheid: BescheidResource = createBescheidResource();
    const file: File = createFile();

    beforeEach(() => {
      service.doUploadBescheidDocument = jest.fn();
    });

    it('should call init upload bescheid document in progress', () => {
      service.uploadBescheidDocument(bescheid, file);

      expect(service.uploadBescheidDocumentInProgress$.value).toEqual({
        loading: true,
        fileName: file.name,
      });
    });

    it('should clear create bescheid document', () => {
      service.createBescheidDocumentInProgress$.next({
        loading: true,
        fileName: 'Dummy',
        error: createApiError(),
      });
      service.uploadBescheidDocument(bescheid, file);

      expect(service.createBescheidDocumentInProgress$.value).toEqual({
        loading: false,
      });
    });

    it('should call do upload bescheid document', () => {
      service.uploadBescheidDocument(bescheid, file);

      expect(service.doUploadBescheidDocument).toHaveBeenCalledWith(bescheid, file);
    });

    it('should return bescheid document file', (done) => {
      service
        .uploadBescheidDocument(bescheid, file)
        .subscribe((uploadFileInProgress: UploadFileInProgress) => {
          expect(uploadFileInProgress.fileName).toBe(file.name);
          expect(uploadFileInProgress.loading).toBeTruthy();
          done();
        });
    });
  });

  describe('do upload bescheid document', () => {
    const bescheid: BescheidResource = createBescheidResource();
    const file: File = createFile();

    const binaryFileStateResource: StateResource<BinaryFileResource> = createStateResource(
      createBinaryFileResource(),
    );

    beforeEach(() => {
      binaryFileService.uploadFile.mockReturnValue(of(binaryFileStateResource));
    });

    it('should call binary file service', () => {
      service.doUploadBescheidDocument(bescheid, file);

      expect(binaryFileService.uploadFile).toHaveBeenCalledWith(
        bescheid,
        BescheidLinkRel.UPLOAD_BESCHEID_FILE,
        file,
        false,
      );
    });

    it('should call handle upload becheid document response', () => {
      service.handleUploadBescheidDocumentResponse = jest.fn();

      service.doUploadBescheidDocument(bescheid, file);

      expect(service.handleUploadBescheidDocumentResponse).toHaveBeenCalledWith(
        bescheid,
        binaryFileStateResource,
      );
    });
  });

  describe('handle upload bescheid document response', () => {
    const bescheid: BescheidResource = createBescheidResource();

    const binaryFile: BinaryFileResource = createBinaryFileResource();
    const apiError: ApiError = createApiError();
    const binaryFileErrorStateResource: StateResource<BinaryFileResource> =
      createErrorStateResource(apiError);
    const binaryFileStateResource: StateResource<BinaryFileResource> =
      createStateResource(binaryFile);

    it('should call create bescheid document from file on success', () => {
      service.createBescheidDocumentFromFile = jest.fn();

      service.handleUploadBescheidDocumentResponse(bescheid, binaryFileStateResource);

      expect(service.createBescheidDocumentFromFile).toHaveBeenCalledWith(bescheid, binaryFile);
    });

    describe('on no error', () => {
      it('should set error', () => {
        service.uploadBescheidDocumentInProgress$.next({ loading: false });

        service.handleUploadBescheidDocumentResponse(bescheid, binaryFileErrorStateResource);

        expect(service.uploadBescheidDocumentInProgress$.value.error).toBe(apiError);
      });

      it('should set loading to false', () => {
        service.uploadBescheidDocumentInProgress$.next({ loading: true });

        service.handleUploadBescheidDocumentResponse(bescheid, binaryFileErrorStateResource);

        expect(service.uploadBescheidDocumentInProgress$.value.loading).toBeFalsy();
      });
    });
  });

  describe('create bescheid document from file', () => {
    const bescheid: BescheidResource = createBescheidResource();
    const binaryFile: BinaryFileResource = createBinaryFileResource();

    const createCommandProps: CreateCommandProps = createCreateCommandProps();
    let buildCreateBescheidDocumentFromFilePropsSpy;

    const commandStateResource: StateResource<CommandResource> =
      createStateResource(createCommandResource());

    beforeEach(() => {
      buildCreateBescheidDocumentFromFilePropsSpy = jest
        .spyOn(BescheidUtil, 'buildCreateBescheidDocumentFromFileProps')
        .mockReturnValue(createCommandProps);
      commandService.createCommandByProps.mockReturnValue(of(commandStateResource));
    });

    it('should call command sevice', () => {
      service.createBescheidDocumentFromFile(bescheid, binaryFile);

      expect(commandService.createCommandByProps).toHaveBeenCalledWith(createCommandProps);
    });

    it('should build create command document from file props', () => {
      service.createBescheidDocumentFromFile(bescheid, binaryFile);

      expect(buildCreateBescheidDocumentFromFilePropsSpy).toHaveBeenCalledWith(
        bescheid,
        binaryFile,
      );
    });

    it('should call handle create bescheid document from file response', () => {
      service.handleCreateBescheidDocumentFromFileResponse = jest.fn();

      service.createBescheidDocumentFromFile(bescheid, binaryFile);

      expect(service.handleCreateBescheidDocumentFromFileResponse).toHaveBeenCalledWith(
        commandStateResource,
        binaryFile,
      );
    });
  });

  describe('handle create bescheid document response from file', () => {
    const binaryFile: BinaryFileResource = createBinaryFileResource();
    const command: CommandResource = createCommandResource([CommandLinkRel.EFFECTED_RESOURCE]);
    const commandStateResource: StateResource<CommandResource> = createStateResource(command);

    beforeEach(() => {
      service.bescheidDocumentFile$.next(createEmptyStateResource());
    });

    it('should set upload bescheid document in progress done', () => {
      service.handleCreateBescheidDocumentFromFileResponse(commandStateResource, binaryFile);

      expect(service.bescheidDocumentFile$.value).toEqual(createStateResource(binaryFile));
    });

    describe('on error', () => {
      it('should set upload bescheid in progress error', () => {
        const httpError: HttpError = createApiError();
        const errorStateResource: StateResource<CommandResource> =
          createErrorStateResource(httpError);

        service.handleCreateBescheidDocumentFromFileResponse(errorStateResource, binaryFile);

        expect(service.uploadBescheidDocumentInProgress$.value.error).toBe(httpError);
      });

      it('should set upload bescheid in progress loading false', () => {
        const httpError: HttpError = createApiError();
        const errorStateResource: StateResource<CommandResource> =
          createErrorStateResource(httpError);

        service.handleCreateBescheidDocumentFromFileResponse(errorStateResource, binaryFile);

        expect(service.uploadBescheidDocumentInProgress$.value.loading).toBeFalsy();
      });
    });

    describe('on success', () => {
      it('should set documentFile', () => {
        service.handleCreateBescheidDocumentFromFileResponse(commandStateResource, binaryFile);

        expect(service.bescheidDocumentFile$.value).toEqual(createStateResource(binaryFile));
      });

      it('should set document uri', () => {
        service.handleCreateBescheidDocumentFromFileResponse(commandStateResource, binaryFile);

        expect(service.bescheidDocumentUri$.value).toEqual(
          getUrl(command, CommandLinkRel.EFFECTED_RESOURCE),
        );
      });
    });
  });

  describe('delete bescheid document', () => {
    it('should clear document uri', () => {
      service.bescheidDocumentUri$.next(faker.internet.url());

      service.deleteBescheidDocument();

      expect(service.bescheidDocumentUri$.value).toBeNull();
    });

    it('should clear document file', () => {
      service.bescheidDocumentFile$.next(createStateResource(createBinaryFileResource()));

      service.deleteBescheidDocument();

      expect(service.bescheidDocumentFile$.value).toEqual(createEmptyStateResource());
    });

    it('should clear document', () => {
      service.bescheidDocument$.next(createStateResource(createDocumentResource()));

      service.deleteBescheidDocument();

      expect(service.bescheidDocument$.value).toEqual(createEmptyStateResource());
    });
  });

  describe('init', () => {
    it('should emit null bescheid document uri', () => {
      service.init();

      expect(service.getDocumentUri()).toBeObservable(singleCold(null));
    });

    it('should emit empty state resource document file', () => {
      service.init();

      expect(service.getBescheidDocumentFile()).toBeObservable(
        singleCold(createEmptyStateResource()),
      );
    });
  });

  describe('create bescheid document', () => {
    const commandResource: CommandResource = createCommandResource([
      CommandLinkRel.EFFECTED_RESOURCE,
    ]);
    const commandStateResource: StateResource<CommandResource> =
      createStateResource(commandResource);

    const bescheidResource: BescheidResource = createBescheidResource();
    const createCommandProps: CreateCommandProps = createCreateCommandProps();

    let buildCreateBescheidDocumentCommandPropsSpy: jest.SpyInstance;

    beforeEach(() => {
      commandService.createCommandByProps.mockReturnValue(of(commandStateResource));
      service.bescheidDraftService.getResource = jest.fn().mockReturnValue(bescheidResource);
      buildCreateBescheidDocumentCommandPropsSpy = jest
        .spyOn(BescheidUtil, 'buildCreateBescheidDocumentCommandProps')
        .mockReturnValue(createCommandProps);
    });

    it('should call command service', () => {
      service.createBescheidDocument();

      expect(commandService.createCommandByProps).toHaveBeenCalledWith(createCommandProps);
    });

    it('should build create command props', () => {
      service.createBescheidDocument();

      expect(buildCreateBescheidDocumentCommandPropsSpy).toHaveBeenCalledWith(bescheidResource);
    });

    it('should set create beschied document in progress', () => {
      service.createBescheidDocument();

      expect(service.createBescheidDocumentInProgress$.value.loading).toBeTruthy();
    });

    it('should set upload beschied document in progress loading false', () => {
      service.uploadBescheidDocumentInProgress$.next({
        loading: true,
        error: createApiError(),
        fileName: 'Dummy',
      });

      service.createBescheidDocument();

      expect(service.uploadBescheidDocumentInProgress$.value).toEqual({ loading: false });
    });

    describe('on successfully done command', () => {
      it('should load bescheid document', fakeAsync(() => {
        service.loadBescheidDocumentByUri = jest.fn();

        service.createBescheidDocument();
        tick();

        expect(service.loadBescheidDocumentByUri).toHaveBeenCalledWith(
          getUrl(commandResource, CommandLinkRel.EFFECTED_RESOURCE),
        );
      }));

      it('should update bescheid document Url', fakeAsync(() => {
        service.loadBescheidDocumentByUri = jest.fn();

        service.createBescheidDocument();
        tick();

        expect(service.bescheidDocumentUri$.value).toBe(
          getUrl(commandResource, CommandLinkRel.EFFECTED_RESOURCE),
        );
      }));
    });

    describe('on error', () => {
      const apiError: HttpError = createApiError();
      const commandErrorStateResource: StateResource<CommandResource> =
        createErrorStateResource(apiError);

      it('should set error ', () => {
        commandService.createCommandByProps.mockReturnValue(of(commandErrorStateResource));

        service.createBescheidDocument();

        expect(service.createBescheidDocumentInProgress$.value.error).toBe(apiError);
      });

      it('should set create bescheid document in progress loading false', () => {
        service.createBescheidDocumentInProgress$.next({ loading: true });
        commandService.createCommandByProps.mockReturnValue(of(commandErrorStateResource));

        service.createBescheidDocument();

        expect(service.createBescheidDocumentInProgress$.value.loading).toBeFalsy();
      });
    });
  });

  describe('exists bescheid document file', () => {
    it('should return true if resource exists', (done) => {
      service.bescheidDocumentFile$.next(createStateResource(createBinaryFileResource()));

      service.existBescheidDocumentFile().subscribe((result) => {
        expect(result).toBeTruthy();
        done();
      });
    });

    it('should return false if resource is null', (done) => {
      service.bescheidDocumentFile$.next(createEmptyStateResource());

      service.existBescheidDocumentFile().subscribe((result) => {
        expect(result).toBeFalsy();
        done();
      });
    });
  });

  describe('get bescheid document command', () => {
    const commandResource: CommandResource = createCommandResource();
    const commandStateResource: StateResource<CommandResource> =
      createStateResource(commandResource);

    beforeEach(() => {
      commandService.getCommandByOrder.mockReturnValue(of(commandStateResource));
    });

    it('should call command service', (done) => {
      service.getBescheidDocumentCommand().subscribe(() => {
        expect(commandService.getCommandByOrder).toHaveBeenCalledWith(
          CommandOrder.CREATE_BESCHEID_DOCUMENT,
        );
        done();
      });
    });

    it('should return command stateResource', (done) => {
      service.getBescheidDocumentCommand().subscribe((result) => {
        expect(result).toBe(commandStateResource);
        done();
      });
    });
  });

  describe('exists bescheid draft', () => {
    it('should call resource service', () => {
      service.bescheidDraftService.exists = jest.fn();

      service.existsBescheidDraft();

      expect(service.bescheidDraftService.exists).toHaveBeenCalled();
    });
  });

  describe('bescheidVerwerfen', () => {
    const command: CommandResource = createCommandResource();
    const commandStateResource: StateResource<CommandResource> = createStateResource(command);

    beforeEach(() => {
      service.deleteBescheid = jest.fn().mockReturnValue(of(commandStateResource));
      service.deleteBescheidDocument = jest.fn();
    });

    it('should get resource', () => {
      service.bescheidDraftService.getResource = jest.fn();

      service.bescheidVerwerfen().pipe(first()).subscribe();

      expect(service.bescheidDraftService.getResource).toHaveBeenCalled();
    });

    it('should delete bescheid', (done) => {
      service.bescheidVerwerfen().subscribe(() => {
        expect(service.deleteBescheid).toHaveBeenCalled();
        done();
      });
    });

    it('should return command', () => {
      const command: StateResource<CommandResource> = createCommandStateResource();
      service.deleteBescheid = jest.fn().mockReturnValue(singleCold(command));

      const command$: Observable<StateResource<CommandResource>> = service.bescheidVerwerfen();

      expect(command$).toBeObservable(singleCold(command));
    });

    it('should delete bescheid document', (done) => {
      service.deleteBescheid = jest.fn().mockReturnValue(of(createCommandStateResource()));

      service.bescheidVerwerfen().subscribe(() => {
        done();
        expect(service.deleteBescheidDocument).toHaveBeenCalled();
      });
    });

    it('should reload vorgang', (done) => {
      service.deleteBescheid = jest.fn().mockReturnValue(of(createCommandStateResource()));

      service.bescheidVerwerfen().subscribe(() => {
        done();
        expect(vorgangService.reloadCurrentVorgang).toHaveBeenCalled();
      });
    });
  });

  describe('load bescheid document file', () => {
    const document: DocumentResource = createDocumentResource([DocumentLinkRel.FILE]);
    const binaryFileStateResource: StateResource<BinaryFileResource> = createStateResource(
      createBinaryFileResource(),
    );

    beforeEach(() => {
      binaryFileService.getFile.mockReturnValue(of(binaryFileStateResource));
    });

    it('should call bianry file service', () => {
      service.loadBescheidDocumentFile(document);

      expect(binaryFileService.getFile).toHaveBeenCalledWith(
        getUrl(document, DocumentLinkRel.FILE),
      );
    });

    it('should set bescheidDocument file', () => {
      service.loadBescheidDocumentFile(document);

      expect(service.bescheidDocumentFile$.value).toBe(binaryFileStateResource);
    });

    it('should set create bescheid document in progress loading false', () => {
      service.createBescheidDocumentInProgress$.next({ loading: true });

      service.loadBescheidDocumentFile(document);

      expect(service.createBescheidDocumentInProgress$.value.loading).toBeFalsy();
    });
  });

  describe('get bescheid list', () => {
    it('should call bescheid list service', () => {
      service.bescheidListService.getList = jest.fn();

      service.getBescheidList();

      expect(service.bescheidListService.getList).toHaveBeenCalled();
    });

    it('should return value', () => {
      const bescheidList: BescheidListResource = createBescheidListResource();
      const bescheidListStateResource: StateResource<BescheidListResource> =
        createStateResource(bescheidList);
      service.bescheidListService.getList = jest
        .fn()
        .mockReturnValue(singleCold(bescheidListStateResource));

      const command$: Observable<StateResource<BescheidListResource>> = service.getBescheidList();

      expect(command$).toBeObservable(singleCold(bescheidListStateResource));
    });
  });

  describe('load bescheid document', () => {
    const resourceUri: ResourceUri = faker.internet.url();
    const document: DocumentResource = createDocumentResource();

    beforeEach(() => {
      resourceRepository.getResource.mockReturnValue(of(document));
    });

    it('should call repository', () => {
      service.loadBescheidDocument(resourceUri);

      expect(resourceRepository.getResource).toHaveBeenCalledWith(resourceUri);
    });

    it('should return value', () => {
      const documentStateResource$: Observable<StateResource<DocumentResource>> =
        service.loadBescheidDocument(resourceUri);

      expect(documentStateResource$).toBeObservable(
        cold('(ab|)', { a: createEmptyStateResource(true), b: createStateResource(document) }),
      );
    });
  });

  describe('getEmpfaenger', () => {
    it('should return Empfänger', () => {
      const vorgangWithEingangResource: VorgangWithEingangResource =
        createVorgangWithEingangResource();
      const vorgangWithEingangStateResource: StateResource<VorgangWithEingangResource> =
        createStateResource(vorgangWithEingangResource);
      vorgangService.getVorgangWithEingang.mockReturnValue(of(vorgangWithEingangStateResource));

      const empfaenger$: Observable<string> = service.getEmpfaenger();

      expect(empfaenger$).toBeObservable(
        singleColdCompleted(
          `${vorgangWithEingangResource.eingang.antragsteller.vorname} ${vorgangWithEingangResource.eingang.antragsteller.nachname}`,
        ),
      );
    });
  });

  describe('get last bescheid', () => {
    const bescheid: BescheidResource = createBescheidResource();
    const bescheide: BescheidResource[] = [bescheid];

    let sortByGermanDateStrSpy: jest.SpyInstance;
    let getItemsSpy: jest.SpyInstance;

    beforeEach(() => {
      service.filterBySentStatus = jest.fn().mockReturnValue(bescheide);
      sortByGermanDateStrSpy = jest
        .spyOn(DateUtil, 'sortByGermanDateStr')
        .mockReturnValue(bescheide);
      getItemsSpy = service.bescheidListService.getItems = jest.fn().mockReturnValue(of(bescheide));
    });

    it('should get items', () => {
      service.getLastBescheid().pipe(first()).subscribe();

      expect(getItemsSpy).toHaveBeenCalled();
    });

    it('should filter by sent status', () => {
      service.getLastBescheid().pipe(first()).subscribe();

      expect(service.filterBySentStatus).toHaveBeenCalledWith(bescheide);
    });

    it('should sort by beschieden am', () => {
      service.getLastBescheid().pipe(first()).subscribe();

      expect(sortByGermanDateStrSpy).toHaveBeenCalledWith(bescheide, expect.any(Function));
    });

    it('should return first element', () => {
      const lastBescheid$: Observable<BescheidResource> = service.getLastBescheid();

      expect(lastBescheid$).toBeObservable(singleColdCompleted(bescheid));
    });
  });

  describe('exist bescheid', () => {
    const bescheid: BescheidResource = createBescheidResource();
    const bescheide: BescheidResource[] = [bescheid];
    const bescheidListStateResource: StateResource<BescheidListResource> = createStateResource(
      createBescheidListResource(bescheide),
    );
    let getItemsSpy: jest.SpyInstance;

    beforeEach(() => {
      service.getBescheidList = jest.fn().mockReturnValue(of(bescheidListStateResource));
      service.filterBySentStatus = jest.fn().mockReturnValue(bescheide);
      getItemsSpy = service.bescheidListService.getItems = jest.fn().mockReturnValue(of(bescheide));
    });

    it('should get items', () => {
      service.existBescheid().pipe(first()).subscribe();

      expect(getItemsSpy).toHaveBeenCalled();
    });

    it('should filter by sent status', () => {
      service.existBescheid().pipe(first()).subscribe();

      expect(service.filterBySentStatus).toHaveBeenCalledWith(bescheide);
    });

    it('should return true if at least one bescheid exists', () => {
      const existBescheid$: Observable<boolean> = service.existBescheid();

      expect(existBescheid$).toBeObservable(singleColdCompleted(true));
    });

    it('should return false if no bescheide exists', () => {
      service.filterBySentStatus = jest.fn().mockReturnValue([]);

      const existBescheid$: Observable<boolean> = service.existBescheid();

      expect(existBescheid$).toBeObservable(singleColdCompleted(false));
    });
  });

  describe('filter by sent status', () => {
    it('should keep entry with sent status', () => {
      const bescheidWithSentStatus: BescheidResource = {
        ...createBescheidResource(),
        status: BescheidStatus.SENT,
      };

      const filteredBescheide: BescheidResource[] = service.filterBySentStatus([
        bescheidWithSentStatus,
      ]);

      expect(filteredBescheide.length).toBe(1);
    });

    it('should filter entry with draft status', () => {
      const bescheidWithDraftStatus: BescheidResource = {
        ...createBescheidResource(),
        status: BescheidStatus.DRAFT,
      };

      const filteredBescheide: BescheidResource[] = service.filterBySentStatus([
        bescheidWithDraftStatus,
      ]);

      expect(filteredBescheide.length).toBe(0);
    });
  });

  describe('refresh list', () => {
    it('should call refresh on list service', () => {
      service.bescheidListService.refresh = jest.fn();

      service.refreshList();

      expect(service.bescheidListService.refresh).toHaveBeenCalled();
    });
  });
});
