/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
import {
  AbstractControl,
  FormControl,
  FormGroup,
  UntypedFormControl,
  UntypedFormGroup,
} from '@angular/forms';
import { createInvalidParam, createIssue } from '../../../test/error';
import { InvalidParam, Issue } from '../tech.model';
import {
  getControlForInvalidParam,
  getControlForIssue,
  getMessageForInvalidParam,
  getMessageForIssue,
  setInvalidParamValidationError,
  setIssueValidationError,
} from './tech.validation.util';

describe('ValidationUtils', () => {
  const baseField1Control: FormControl = new UntypedFormControl();
  const baseField2Control: FormControl = new UntypedFormControl();
  const subGroupFieldControl: FormControl = new UntypedFormControl();

  const form: FormGroup = new UntypedFormGroup({
    baseField1: baseField1Control,
    baseField2: baseField2Control,
    subGroup: new UntypedFormGroup({
      subGroupField1: subGroupFieldControl,
    }),
  });

  describe('set issue validation error', () => {
    describe('get control for issue', () => {
      it('should return base field control', () => {
        const issue: Issue = { ...createIssue(), field: 'baseField1' };

        const control: AbstractControl = getControlForIssue(form, issue);

        expect(control).toBe(baseField1Control);
      });

      it('should return sub group field', () => {
        const issue: Issue = { ...createIssue(), field: 'subGroup.subGroupField1' };

        const control: AbstractControl = getControlForIssue(form, issue);

        expect(control).toBe(subGroupFieldControl);
      });

      it('should ignore path prefix', () => {
        const issue: Issue = { ...createIssue(), field: 'pathprefix.resource.baseField1' };

        const control: AbstractControl = getControlForIssue(form, issue, 'pathprefix.resource');

        expect(control).toBe(baseField1Control);
      });
    });

    describe('in base field', () => {
      const issue: Issue = { ...createIssue(), field: 'baseField1' };

      it('should set error in control', () => {
        setIssueValidationError(form, issue);

        expect(baseField1Control.errors).not.toBeNull();
      });

      it('should set message code in control', () => {
        setIssueValidationError(form, issue);

        expect(baseField1Control.hasError(issue.messageCode)).toBe(true);
      });

      it('should set control touched', () => {
        setIssueValidationError(form, issue);

        expect(baseField1Control.touched).toBe(true);
      });

      it('should not set error in other control', () => {
        setIssueValidationError(form, issue);

        expect(baseField2Control.errors).toBeNull();
      });
    });

    describe('in subGroup Field', () => {
      const issue: Issue = { ...createIssue(), field: 'subGroup.subGroupField1' };

      it('should set error in control', () => {
        setIssueValidationError(form, issue);

        expect(subGroupFieldControl.errors).not.toBeNull();
      });
    });
  });

  describe('get message for issue', () => {
    const fieldLabel: string = 'Field Label';

    it('should return message', () => {
      const msg: string = getMessageForIssue(fieldLabel, {
        ...createIssue(),
        messageCode: 'validation_field_size',
      });

      expect(msg).toContain('muss mindestens');
    });

    it('should set field label', () => {
      const msg: string = getMessageForIssue(fieldLabel, {
        ...createIssue(),
        messageCode: 'validation_field_size',
      });

      expect(msg).toContain(fieldLabel);
    });

    it('should replace min param', () => {
      const msg: string = getMessageForIssue(fieldLabel, {
        ...createIssue(),
        messageCode: 'validation_field_size',
        parameters: [{ name: 'min', value: '3' }],
      });

      expect(msg).toContain('3');
    });
  });

  describe('invalid param', () => {
    const formPrefixes: string[] = ['', 'some-prefix'];
    const fieldNames: string[] = ['baseField1', 'baseField2', 'subGroup.subGroupField1'];
    const prefixNameCombinations: string[][] = formPrefixes.flatMap((prefix) =>
      fieldNames.map((name) => [prefix, name]),
    );
    const unknownName = 'unknown-field';

    describe.each(prefixNameCombinations)(
      'with prefix "%s" and fieldName "%s"',
      (prefix, fieldName) => {
        let invalidParam: InvalidParam;

        beforeEach(() => {
          form.reset();
          invalidParam = {
            ...createInvalidParam(),
            name: prefix.length ? `${prefix}.${fieldName}` : fieldName,
          };
        });

        describe('get message for invalid param', () => {
          it('should return', () => {
            const msg: string = getMessageForInvalidParam(invalidParam, prefix);

            expect(msg).toEqual(`Bitte ${fieldName} ausfüllen`);
          });
        });

        describe('get control for invalid param', () => {
          it('should find', () => {
            const control: AbstractControl = getControlForInvalidParam(form, invalidParam, prefix);

            expect(control).toBeTruthy();
          });
        });

        describe('set invalid param validation error', () => {
          it('should assign invalidParam to form control error without prefix', () => {
            const message: string = getMessageForInvalidParam(invalidParam, prefix);

            setInvalidParamValidationError(form, invalidParam, prefix);

            const errorMessage: string = form.getError(invalidParam.reason, fieldName);
            expect(errorMessage).toBe(message);
          });

          it('should mark form as touched', () => {
            setInvalidParamValidationError(form, invalidParam, prefix);

            expect(form.touched).toBeTruthy();
          });
        });
      },
    );

    describe.each([
      ['', '', 'unknown-field'],
      ['valid-prefix', 'valid-prefix', 'unknown-field'],
      ['valid-prefix', 'valid-prefix', 'subGroup.unknown-field'],
      ['unknown-prefix', 'valid-prefix', 'unknown-field'],
      ['unknown-prefix', 'valid-prefix', 'baseField1'],
    ])(
      'with pathPrefix "%s", paramPrefix "%s", and field-name "%s"',
      (pathPrefix, paramPrefix, fieldName) => {
        let invalidParam: InvalidParam;

        beforeEach(() => {
          form.reset();
          invalidParam = createInvalidParam();
          invalidParam.name = paramPrefix.length > 0 ? `${paramPrefix}.${fieldName}` : fieldName;
        });

        it('should not find form control', () => {
          const control: AbstractControl = getControlForInvalidParam(
            form,
            invalidParam,
            pathPrefix,
          );

          expect(control).toBeFalsy();
        });

        it('should not assign to field control error', () => {
          setInvalidParamValidationError(form, invalidParam, pathPrefix);

          const errorMessage = form.getError(invalidParam.reason, unknownName);
          expect(errorMessage).toBeFalsy();
        });

        it('should not mark as touched', () => {
          setInvalidParamValidationError(form, invalidParam, pathPrefix);

          expect(form.touched).toBeFalsy();
        });
      },
    );
  });
});
