import { BescheidLinkRel, BescheidResource, BescheidService } from '@alfa-client/bescheid-shared';
import { BinaryFileAttachmentContainerComponent } from '@alfa-client/binary-file';
import { BinaryFileResource, BinaryFileService } from '@alfa-client/binary-file-shared';
import { createEmptyStateResource, createStateResource } from '@alfa-client/tech-shared';
import { Mock, mock, useFromMock } from '@alfa-client/test-utils';
import { ComponentFixture, TestBed } from '@angular/core/testing';
import { ReactiveFormsModule, UntypedFormBuilder } from '@angular/forms';
import { FileUploadEditorComponent } from '@ods/component';
import { cold } from 'jest-marbles';
import { MockComponent } from 'ng-mocks';
import { EMPTY, Observable, Subscription, of } from 'rxjs';
import { createBescheidResource } from '../../../../../../../../../bescheid-shared/src/test/bescheid';
import { createBinaryFileResource } from '../../../../../../../../../binary-file-shared/test/binary-file';
import { BescheidenFormService } from '../../../../bescheiden.formservice';
import { VorgangDetailBescheidenAttachmentHochladenComponent } from './vorgang-detail-bescheiden-attachment-hochladen.component';

describe('VorgangDetailBescheidenDokumentHochladenComponent', () => {
  let component: VorgangDetailBescheidenAttachmentHochladenComponent;
  let fixture: ComponentFixture<VorgangDetailBescheidenAttachmentHochladenComponent>;
  const selfLink: string = 'self';

  let bescheidService: Mock<BescheidService>;
  let binaryFileService: Mock<BinaryFileService>;
  let formService: BescheidenFormService;

  beforeEach(async () => {
    bescheidService = mock(BescheidService);
    bescheidService.getAttachments.mockReturnValue(EMPTY);

    binaryFileService = mock(BinaryFileService);

    formService = new BescheidenFormService(new UntypedFormBuilder(), useFromMock(bescheidService));

    await TestBed.configureTestingModule({
      declarations: [
        VorgangDetailBescheidenAttachmentHochladenComponent,
        MockComponent(BinaryFileAttachmentContainerComponent),
        MockComponent(FileUploadEditorComponent),
      ],
      imports: [ReactiveFormsModule],
      providers: [
        {
          provide: BescheidenFormService,
          useValue: formService,
        },
        {
          provide: BescheidService,
          useValue: bescheidService,
        },
        {
          provide: BinaryFileService,
          useValue: binaryFileService,
        },
      ],
    }).compileComponents();

    fixture = TestBed.createComponent(VorgangDetailBescheidenAttachmentHochladenComponent);
    component = fixture.componentInstance;
    fixture.detectChanges();
  });

  it('should create', () => {
    expect(component).toBeTruthy();
  });

  describe('ngOnInit', () => {
    let loadExistingAttachments: jest.Mock;
    let subscribeToFileDeletion: jest.Mock;

    beforeEach(() => {
      loadExistingAttachments = component.loadExistingAttachments = jest.fn();
      subscribeToFileDeletion = component.subscribeToFileDeletion = jest.fn();
    });

    it('should load existing attachments', () => {
      component.ngOnInit();

      expect(loadExistingAttachments).toHaveBeenCalled();
    });

    it('should subscribe to file deletion', () => {
      component.ngOnInit();

      expect(subscribeToFileDeletion).toHaveBeenCalled();
    });
  });

  describe('ngOnDestroy', () => {
    it('should unsubscribe from file deletion', () => {
      const subscription = mock(Subscription);
      const getDeleteFile = (formService.getFileDelete = jest.fn());
      const observable = mock(Observable);
      getDeleteFile.mockReturnValue(observable);
      observable.subscribe.mockReturnValue(subscription);
      component.ngOnInit();

      component.ngOnDestroy();

      expect(subscription.unsubscribe).toHaveBeenCalled();
    });
  });

  describe('loadExistingAttachments', () => {
    let binaryFileResource: BinaryFileResource;

    beforeEach(() => {
      binaryFileResource = createBinaryFileResource();
      bescheidService.getAttachments.mockReturnValue(of([binaryFileResource]));
    });

    it('should get attachments', () => {
      component.loadExistingAttachments();

      expect(bescheidService.getAttachments).toHaveBeenCalled();
    });

    it('should add attachments to fileList', () => {
      component.loadExistingAttachments();

      expect(component.fileList).toEqual([binaryFileResource]);
    });
  });

  describe('subscribeToFileDeletion', () => {
    let getFileDelete: jest.Mock;
    let getFileDeleteObservable: Mock<Observable<BinaryFileResource>>;

    beforeEach(() => {
      getFileDeleteObservable = mock(Observable);
      getFileDelete = formService.getFileDelete = jest
        .fn()
        .mockReturnValue(getFileDeleteObservable);
    });

    it('should subscribe to file deletion', () => {
      component.subscribeToFileDeletion();

      expect(getFileDelete).toHaveBeenCalled();
    });
  });

  describe('deleteFile', () => {
    let binaryFileResource: BinaryFileResource;

    beforeEach(() => {
      binaryFileResource = createBinaryFileResource();
    });

    it('should remove file from list', () => {
      component.fileList = [binaryFileResource];

      component.deleteFile(binaryFileResource);

      expect(component.fileList).toEqual([]);
    });
  });

  describe('getFileLinkList', () => {
    it('should return url list', () => {
      const binaryFileResource = createBinaryFileResource([selfLink]);
      component.fileList = [binaryFileResource];

      const uriList = component.getFileLinkList();

      expect(uriList[0]).toEqual(binaryFileResource._links.self.href);
    });
  });

  describe('uploadFile', () => {
    const file: File = <any>{ name: 'TestDatei' };
    let uploadAndGetFile: jest.Mock;

    beforeEach(() => {
      uploadAndGetFile = component.uploadAndGetFile = jest.fn();
    });

    it('should uploadAndGetFile', () => {
      uploadAndGetFile.mockReturnValue(of(createEmptyStateResource()));

      component.uploadFile(file);

      expect(component.uploadAndGetFile).toHaveBeenCalledWith(file);
    });

    it('should start with loading empty state resource', () => {
      const loadingStateResource = createEmptyStateResource(true);
      const uploadStateResource = createStateResource(createBinaryFileResource());
      uploadAndGetFile.mockReturnValue(cold('-a', { a: uploadStateResource }));

      component.uploadFile(file);

      expect(component.uploadInProgress$).toBeObservable(
        cold('ab', { a: loadingStateResource, b: uploadStateResource }),
      );
    });

    it('should call form service', () => {
      formService.uploadAttachment = jest.fn();
      uploadAndGetFile.mockReturnValue(of(createEmptyStateResource()));

      component.uploadFile(file);
      component.uploadInProgress$.subscribe();

      expect(formService.uploadAttachment).toHaveBeenCalledWith(createEmptyStateResource());
    });
  });

  describe('uploadAndGetFile', () => {
    const file: File = <any>{ name: 'TestDatei' };
    let bescheidResource: BescheidResource;
    let uploadResource: BinaryFileResource;

    beforeEach(() => {
      bescheidResource = createBescheidResource();
      uploadResource = createBinaryFileResource();
      bescheidService.getBescheidDraft.mockReturnValue(of(createStateResource(bescheidResource)));
      binaryFileService.uploadFile.mockReturnValue(of(createStateResource(uploadResource)));
    });

    it('should get bescheid draft', () => {
      component.uploadAndGetFile(file);

      expect(bescheidService.getBescheidDraft).toHaveBeenCalled();
    });

    it('should upload file', (done) => {
      const binaryFileStateResource$ = component.uploadAndGetFile(file);

      binaryFileStateResource$.subscribe(() => {
        expect(binaryFileService.uploadFile).toHaveBeenCalledWith(
          bescheidResource,
          BescheidLinkRel.UPLOAD_ATTACHMENT,
          file,
          false,
        );
        done();
      });
    });

    it('should add file to file list', (done) => {
      const binaryFileStateResource$ = component.uploadAndGetFile(file);

      binaryFileStateResource$.subscribe(() => {
        expect(component.fileList[0]).toEqual(uploadResource);
        done();
      });
    });

    it('should emit binary file state resource', () => {
      const binaryFileStateResource$ = component.uploadAndGetFile(file);

      expect(binaryFileStateResource$).toBeObservable(
        cold('(a|)', { a: createStateResource(uploadResource) }),
      );
    });
  });
});
