package de.ozgcloud.alfa.bescheid;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.time.ZonedDateTime;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.UUID;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.EnumSource;
import org.junit.jupiter.params.provider.EnumSource.Mode;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.alfa.common.attacheditem.VorgangAttachedItem;
import de.ozgcloud.alfa.common.attacheditem.VorgangAttachedItemService;
import de.ozgcloud.alfa.common.attacheditem.VorgangAttachedItemTestFactory;
import de.ozgcloud.alfa.common.command.Command;
import de.ozgcloud.alfa.common.command.CommandOrder;
import de.ozgcloud.alfa.common.command.CommandTestFactory;

class BescheidHistorieProcessorTest {

	@Spy
	@InjectMocks
	private BescheidHistorieProcessor processor;
	@Mock
	private VorgangAttachedItemService vorgangAttachedItemService;

	@DisplayName("Process")
	@Nested
	class TestProcess {

		private final Command command = CommandTestFactory.create();
		private final List<Command> commands = Collections.singletonList(command);

		@BeforeEach
		void mock() {
			doReturn(commands).when(processor).filterByOrder(commands);
			doReturn(commands).when(processor).filterBescheidRelatedCommands(commands);
		}

		@Test
		void shouldFilterByOrder() {
			processor.process(commands);

			verify(processor).filterByOrder(commands);
		}

		@Test
		void shouldFilterBescheidRelatedCommands() {
			processor.process(commands);

			verify(processor).filterBescheidRelatedCommands(commands);
		}

		@Test
		void shouldReturnCommands() {
			var processedCommands = processor.process(commands);

			assertThat(processedCommands).containsExactly(command);
		}
	}

	@DisplayName("Filter by order")
	@Nested
	class TestFilterByOrder {

		@ParameterizedTest
		@EnumSource(mode = Mode.INCLUDE, names = { "CREATE_BESCHEID", "DELETE_BESCHEID" })
		void shouldFilterCommandWithOrder(CommandOrder order) {
			var command = CommandTestFactory.createBuilder().order(order.toString()).build();

			var commands = processor.filterByOrder(Collections.singletonList(command));

			assertThat(commands).isEmpty();
		}

		@ParameterizedTest
		@EnumSource(mode = Mode.EXCLUDE, names = { "CREATE_BESCHEID", "DELETE_BESCHEID" })
		void shouldKeepCommandWithOrder(CommandOrder order) {
			var command = CommandTestFactory.createBuilder().order(order.toString()).build();

			var commands = processor.filterByOrder(Collections.singletonList(command));

			assertThat(commands).containsExactly(command);
		}

		@DisplayName("should filter bescheid related create attached item command")
		@Test
		void shouldFilterCreateAttachedItemCommand() {
			var command = CommandTestFactory.createBuilder().order(CommandOrder.CREATE_ATTACHED_ITEM.toString())
					.body(Map.of(VorgangAttachedItem.FIELD_ITEM_NAME, BescheidHistorieProcessor.BESCHEID_ITEM_NAME))
					.build();

			var commands = processor.filterByOrder(Collections.singletonList(command));

			assertThat(commands).isEmpty();
		}

		@DisplayName("should filter bescheid related update attached item command")
		@Test
		void shouldFilterUpdateAttachedItemCommand() {
			var command = CommandTestFactory.createBuilder().order(CommandOrder.UPDATE_ATTACHED_ITEM.toString())
					.body(Map.of(VorgangAttachedItem.FIELD_ITEM_NAME, BescheidHistorieProcessor.BESCHEID_ITEM_NAME))
					.build();

			var commands = processor.filterByOrder(Collections.singletonList(command));

			assertThat(commands).isEmpty();
		}
	}

	@DisplayName("Filter bescheid related commands")
	@Nested
	class TestFilterBescheidRelatedCommands {

		private final Command commandToFilter = CommandTestFactory.createBuilder().id(UUID.randomUUID().toString()).build();
		private final Command command = CommandTestFactory.create();
		private final List<Command> commands = List.of(commandToFilter, command);

		@BeforeEach
		void mock() {
			doReturn(Collections.singletonList(commandToFilter.getId())).when(processor).getBescheidRelatedCommandIdsToFilter(any());
		}

		@Test
		void shouldGetBescheidRelatedCommandIdsToFilter() {
			processor.filterBescheidRelatedCommands(commands);

			verify(processor).getBescheidRelatedCommandIdsToFilter(commands);
		}

		@Test
		void shouldReturnFilteredList() {
			var filteredCommands = processor.filterBescheidRelatedCommands(commands);

			assertThat(filteredCommands).containsExactly(command);
		}
	}

	@DisplayName("Get bescheid related command ids to filter")
	@Nested
	class TestGetBescheidRelatedCommandIdsToFilter {

		private final Command sendBescheidCommand = CommandTestFactory.createBuilder()
				.id(UUID.randomUUID().toString())
				.relationId(BescheidTestFactory.ID)
				.order(CommandOrder.SEND_BESCHEID.toString())
				.build();
		private final Command updateBescheidCommand = CommandTestFactory.createBuilder()
				.id(UUID.randomUUID().toString())
				.relationId(BescheidTestFactory.ID)
				.order(CommandOrder.UPDATE_BESCHEID.toString())
				.build();
		private final Command updateBescheidCommandOfBescheidNotSent = CommandTestFactory.createBuilder()
				.id(UUID.randomUUID().toString())
				.relationId(UUID.randomUUID().toString())
				.order(CommandOrder.UPDATE_BESCHEID.toString())
				.build();
		private final List<Command> commands = List.of(sendBescheidCommand, updateBescheidCommand, updateBescheidCommandOfBescheidNotSent);

		@BeforeEach
		void mock() {
			doReturn(Collections.singletonList(updateBescheidCommand.getId())).when(processor).getCommandIdsToFilter(any(), any());
		}

		@Test
		void shouldGetCommandIdsToFilter() {
			processor.getBescheidRelatedCommandIdsToFilter(commands);

			verify(processor).getCommandIdsToFilter(sendBescheidCommand, commands);
		}

		@Test
		void shouldReturnIdsToFilter() {
			var commandIds = processor.getBescheidRelatedCommandIdsToFilter(commands);

			assertThat(commandIds).containsExactlyInAnyOrder(updateBescheidCommand.getId(), updateBescheidCommandOfBescheidNotSent.getId());
		}
	}

	@DisplayName("Get commands to filter")
	@Nested
	class TestGetCommandsToFilter {

		private final Command sendBescheidCommand = CommandTestFactory.createBuilder()
				.id(UUID.randomUUID().toString())
				.relationId(BescheidTestFactory.ID)
				.order(CommandOrder.SEND_BESCHEID.toString()).build();

		private final Command relatedUpdateBescheidCommand = CommandTestFactory.createBuilder().id(UUID.randomUUID().toString())
				.createdAt(ZonedDateTime.now()).build();
		private final Command relatedLatestUpdateBescheidCommand = CommandTestFactory.createBuilder().id(UUID.randomUUID().toString())
				.createdAt(ZonedDateTime.now().plusMinutes(10))
				.build();
		private final List<Command> relatedUpdateBescheidCommands = List.of(relatedUpdateBescheidCommand, relatedLatestUpdateBescheidCommand);

		private final Map<String, ?> bescheidItemMap = BescheidTestFactory.createAsMap();
		private final VorgangAttachedItem vorgangAttachedItem = VorgangAttachedItemTestFactory.createBuilder().item(bescheidItemMap).build();

		@BeforeEach
		void mock() {
			when(vorgangAttachedItemService.getById(anyString())).thenReturn(vorgangAttachedItem);
			doReturn(relatedUpdateBescheidCommands).when(processor).getRelatedUpdateBescheidCommands(sendBescheidCommand,
					relatedUpdateBescheidCommands);
		}

		@Test
		void shouldGetRelatedUpdateBescheidCommands() {
			processor.getCommandIdsToFilter(sendBescheidCommand, relatedUpdateBescheidCommands);

			verify(processor).getRelatedUpdateBescheidCommands(sendBescheidCommand, relatedUpdateBescheidCommands);
		}

		@Test
		void shouldGetVorgangAttachedItem() {
			processor.getCommandIdsToFilter(sendBescheidCommand, relatedUpdateBescheidCommands);

			verify(processor).getVorgangAttachedItem(sendBescheidCommand.getRelationId());
		}

		@Test
		void shouldCheckIfManualSent() {
			processor.getCommandIdsToFilter(sendBescheidCommand, relatedUpdateBescheidCommands);

			verify(processor).isManualSent(bescheidItemMap);
		}

		@DisplayName("on send by manual")
		@Nested
		class TestOnSendByManualCommand {

			private final VorgangAttachedItem vorgangAttachedItem = createVorgangAttachedItem(createBescheidMap(SendBy.MANUAL));

			@BeforeEach
			void mock() {
				when(vorgangAttachedItemService.getById(anyString())).thenReturn(vorgangAttachedItem);
				doReturn(Stream.of(relatedUpdateBescheidCommand.getId())).when(processor).filterLatestByCreatedAt(any());
			}

			@Test
			void shouldCallFilterLatestByCreatedAt() {
				processor.getCommandIdsToFilter(sendBescheidCommand, relatedUpdateBescheidCommands);

				verify(processor).filterLatestByCreatedAt(relatedUpdateBescheidCommands);
			}

			@Test
			void shouldReturnRelatedCommandIds() {
				var ids = processor.getCommandIdsToFilter(sendBescheidCommand, relatedUpdateBescheidCommands);

				assertThat(ids).containsExactly(sendBescheidCommand.getId(), relatedUpdateBescheidCommand.getId());
			}
		}

		@DisplayName("on send by nachricht")
		@Nested
		class TestOnSendByNachrichtCommand {

			private final VorgangAttachedItem vorgangAttachedItem = createVorgangAttachedItem(createBescheidMap(SendBy.NACHRICHT));

			@BeforeEach
			void mock() {
				when(vorgangAttachedItemService.getById(anyString())).thenReturn(vorgangAttachedItem);
			}

			@Test
			void shouldReturnRelatedCommandIds() {
				var ids = processor.getCommandIdsToFilter(sendBescheidCommand, relatedUpdateBescheidCommands);

				assertThat(ids).containsExactly(relatedUpdateBescheidCommand.getId(), relatedLatestUpdateBescheidCommand.getId());
			}
		}

		private VorgangAttachedItem createVorgangAttachedItem(Map<String, Object> item) {
			return VorgangAttachedItemTestFactory.createBuilder()
					.id(BescheidTestFactory.ID)
					.itemName(BescheidHistorieProcessor.BESCHEID_ITEM_NAME)
					.item(item)
					.build();
		}

		private Map<String, Object> createBescheidMap(SendBy sendBy) {
			var item = BescheidTestFactory.createAsMap();
			item.remove(Bescheid.SEND_BY_FIELD);
			item.put(Bescheid.SEND_BY_FIELD, sendBy);
			return item;
		}
	}

	@DisplayName("Get related update bescheid commands")
	@Nested
	class TestGetRelatedUpdateBescheiCommands {

		private final Command command = CommandTestFactory.create();

		@Test
		void shouldFilterOnMismatchRelationId() {
			var commandWithNonMatchingRelationId = CommandTestFactory.createBuilder().relationId(UUID.randomUUID().toString()).build();

			var relatedCommands = processor.getRelatedUpdateBescheidCommands(command, Collections.singletonList(commandWithNonMatchingRelationId));

			assertThat(relatedCommands).isEmpty();
		}

		@ParameterizedTest
		@EnumSource(mode = Mode.INCLUDE, names = { "UPDATE_BESCHEID" })
		void shouldKeepUpdateBescheidCommandsOnly(CommandOrder order) {
			var commandWithNonMatchingOrder = CommandTestFactory.createBuilder().order(order.toString()).build();

			var relatedCommands = processor.getRelatedUpdateBescheidCommands(command, Collections.singletonList(commandWithNonMatchingOrder));

			assertThat(relatedCommands).containsExactly(commandWithNonMatchingOrder);
		}

		@ParameterizedTest
		@EnumSource(mode = Mode.EXCLUDE, names = { "UPDATE_BESCHEID" })
		void shouldFilterNonUpdateBescheidCommands(CommandOrder order) {
			var commandWithNonMatchingOrder = CommandTestFactory.createBuilder().order(order.toString()).build();

			var relatedCommands = processor.getRelatedUpdateBescheidCommands(command, Collections.singletonList(commandWithNonMatchingOrder));

			assertThat(relatedCommands).isEmpty();
		}
	}

	@DisplayName("Get vorgangAttachedItem")
	@Nested
	class TestGetVorgangAttachedItem {

		private final VorgangAttachedItem vorgangAttachedItem = VorgangAttachedItemTestFactory.create();

		@BeforeEach
		void mock() {
			when(vorgangAttachedItemService.getById(anyString())).thenReturn(vorgangAttachedItem);
		}

		@Test
		void shouldCallService() {
			processor.getVorgangAttachedItem(VorgangAttachedItemTestFactory.ID);

			verify(vorgangAttachedItemService).getById(VorgangAttachedItemTestFactory.ID);
		}

		@Test
		void shouldReturn() {
			var vorgangAttachedItem = processor.getVorgangAttachedItem(VorgangAttachedItemTestFactory.ID);

			assertThat(vorgangAttachedItem).isNotNull().usingRecursiveComparison().isEqualTo(vorgangAttachedItem);
		}
	}

	@DisplayName("Is manual sent")
	@Nested
	class TestIsManualSent {

		@Test
		void shouldReturnTrueOnManual() {
			var manualSent = processor.isManualSent(Map.of(Bescheid.SEND_BY_FIELD, SendBy.MANUAL));

			assertThat(manualSent).isTrue();
		}

		@Test
		void shouldReturnFalseOnNachricht() {
			var manualSent = processor.isManualSent(Map.of(Bescheid.SEND_BY_FIELD, SendBy.NACHRICHT));

			assertThat(manualSent).isFalse();
		}

		@Test
		void shouldThrowOnMissingSendBy() {
			assertThatThrownBy(() -> processor.isManualSent(Map.of())).isInstanceOf(NullPointerException.class);
		}
	}

	@DisplayName("Filter latest by createdAt")
	@Nested
	class TestFilterLatestByCreatedAt {

		private final Command command1 = CommandTestFactory.createBuilder().id(UUID.randomUUID().toString())
				.createdAt(ZonedDateTime.now().minusHours(12)).build();
		private final Command command2 = CommandTestFactory.createBuilder().id(UUID.randomUUID().toString()).createdAt(ZonedDateTime.now()).build();
		private final Command command3 = CommandTestFactory.createBuilder().id(UUID.randomUUID().toString())
				.createdAt(ZonedDateTime.now().plusHours(12)).build();

		private final List<Command> commands = List.of(command2, command3, command1);

		@Test
		void shouldFilterLatestCreatedAtCommand() {
			var filteredCommands = processor.filterLatestByCreatedAt(commands);

			assertThat(filteredCommands).containsExactlyInAnyOrder(command1.getId(), command2.getId());
		}

		@Test
		void shouldReturnEmptyIfThereAreNoCommands() {
			var latestCommandId = processor.filterLatestByCreatedAt(List.of());

			assertThat(latestCommandId).isEmpty();
		}
	}
}