package de.ozgcloud.alfa.bescheid;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Optional;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.ValueSource;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.alfa.vorgang.VorgangHeaderTestFactory;

class BescheidServiceTest {

	@Spy
	@InjectMocks
	private BescheidService service;
	@Mock
	private BescheidRemoteService remoteService;

	@Nested
	class TestGetBescheidDraft {

		@Test
		void shouldCallRemoteService() {
			service.getBescheidDraft(VorgangHeaderTestFactory.ID);

			verify(remoteService).getBescheidDraft(VorgangHeaderTestFactory.ID);
		}

		@Test
		void shouldReturnResultFromRemoteService() {
			var remoteServiceResult = Optional.of(BescheidTestFactory.create());
			when(remoteService.getBescheidDraft(VorgangHeaderTestFactory.ID)).thenReturn(remoteServiceResult);

			var result = service.getBescheidDraft(VorgangHeaderTestFactory.ID);

			assertThat(result).isEqualTo(remoteServiceResult);
		}
	}

	@Nested
	class TestFindByVorgangId {
		private final Bescheid bescheid = BescheidTestFactory.create();

		@BeforeEach
		void mockRemoteService() {
			when(remoteService.findByVorgangId(VorgangHeaderTestFactory.ID)).thenReturn(Stream.of(bescheid));
		}

		@Test
		void shouldCallRemoteService() {
			callService();

			verify(remoteService).findByVorgangId(VorgangHeaderTestFactory.ID);
		}

		@Test
		void shouldReturnStream() {
			var resultStream = callService();

			assertThat(resultStream).containsExactly(bescheid);
		}

		private Stream<Bescheid> callService() {
			return remoteService.findByVorgangId(VorgangHeaderTestFactory.ID);
		}
	}

	@Nested
	class TestCanCreateBescheidDocumentAutomatically {

		@Test
		void shouldCallRemoteService() {
			service.canCreateBescheidDocumentAutomatically();

			verify(remoteService).canCreateBescheidDocument();
		}

		@ParameterizedTest
		@ValueSource(booleans = { true, false })
		void shouldRetrun(boolean canCreateBescheidDocument) {
			when(remoteService.canCreateBescheidDocument()).thenReturn(canCreateBescheidDocument);

			var canCreateAutomatically = service.canCreateBescheidDocumentAutomatically();

			assertThat(canCreateAutomatically).isEqualTo(canCreateBescheidDocument);
		}
	}

	@DisplayName("Exists bescheid")
	@Nested
	class TestExistsBescheid {

		@Test
		void shouldCallRemoteService() {
			service.existsBescheid(VorgangHeaderTestFactory.ID);

			verify(remoteService).findByVorgangId(VorgangHeaderTestFactory.ID);
		}

		@DisplayName("on existing bescheid")
		@Nested
		class TestOnExistingBescheid {

			@BeforeEach
			void mock() {
				when(remoteService.findByVorgangId(anyString())).thenReturn(Stream.of(BescheidTestFactory.create()));
			}

			@Test
			void shouldReturnTrue() {
				var exists = service.existsBescheid(VorgangHeaderTestFactory.ID);

				assertThat(exists).isTrue();
			}
		}

		@DisplayName("on missing bescheid")
		@Nested
		class TestOnMissingBescheid {

			@BeforeEach
			void mock() {
				when(remoteService.findByVorgangId(anyString())).thenReturn(Stream.empty());
			}

			@Test
			void shouldReturnFalse() {
				var exists = service.existsBescheid(VorgangHeaderTestFactory.ID);

				assertThat(exists).isFalse();
			}
		}
	}
}
