package de.ozgcloud.alfa.export;

import static de.ozgcloud.alfa.common.TestUtils.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.util.List;
import java.util.UUID;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;

import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.MockedStatic;
import org.mockito.Spy;

import de.ozgcloud.alfa.common.ExportFilenameGenerator;
import de.ozgcloud.alfa.common.HistorienProtokollInformationTypeTestFactory;
import de.ozgcloud.alfa.common.TestUtils;
import de.ozgcloud.alfa.common.file.OzgFile;
import de.ozgcloud.alfa.common.file.OzgFileTestFactory;
import de.ozgcloud.alfa.file.ExportFileService;
import de.ozgcloud.alfa.historie.ExportHistorieService;
import de.ozgcloud.alfa.kommentar.ExportKommentarService;
import de.ozgcloud.alfa.kommentar.KommentarsExportData;
import de.ozgcloud.alfa.kommentar.KommentarsExportDataTestFactory;
import de.ozgcloud.alfa.postfach.ExportNachrichtService;
import de.ozgcloud.alfa.postfach.PostfachMailExportData;
import de.ozgcloud.alfa.postfach.PostfachMailExportDataTestFactory;
import de.ozgcloud.alfa.vorgang.EingangHeaderTestFactory;
import de.ozgcloud.alfa.vorgang.EingangTestFactory;
import de.ozgcloud.alfa.vorgang.ExportVorgangService;
import de.ozgcloud.alfa.vorgang.VorgangHeaderTestFactory;
import de.ozgcloud.alfa.vorgang.VorgangTypeTestFactory;
import de.ozgcloud.alfa.vorgang.VorgangWithEingang;
import de.ozgcloud.alfa.vorgang.VorgangWithEingangTestFactory;
import de.ozgcloud.common.errorhandling.TechnicalException;
import de.xoev.xdomea.AbgabeAbgabe0401;
import de.xoev.xdomea.AkteType;
import de.xoev.xdomea.DokumentType;
import de.xoev.xdomea.HistorienProtokollInformationType;
import de.xoev.xdomea.NkAbgabeType;
import de.xoev.xdomea.VorgangType;
import lombok.SneakyThrows;

class ExportServiceTest {

	@Spy
	@InjectMocks
	private ExportService service;

	@Mock
	private XdomeaXmlMarshaller xdomeaXmlMarshaller;

	@Mock
	private ExportFileService exportFileService;
	@Mock
	private ExportVorgangService exportVorgangService;
	@Mock
	private ExportHistorieService exportHistorieService;
	@Mock
	private ExportKommentarService exportKommentarService;
	@Mock
	private ExportNachrichtService exportNachrichtService;

	@DisplayName("Write exportToXdomea")
	@Nested
	class TestWriteExport {

		private static final String FILENAME_ID = UUID.randomUUID().toString();
		private final ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
		private final ExportData exportData = ExportDataTestFactory.create();

		@Mock
		private File zipFile;

		@BeforeEach
		void setUp() {
			doReturn(exportData).when(service).collectExportData(VorgangHeaderTestFactory.ID, FILENAME_ID);
			doReturn(zipFile).when(service).createZipFile(exportData);
			doNothing().when(service).writeZipFileContent(zipFile, outputStream);
		}

		@Test
		void shouldCollectExportData() {
			callService();

			verify(service).collectExportData(VorgangHeaderTestFactory.ID, FILENAME_ID);
		}

		@Test
		void shouldCreateZipFile() {
			callService();

			verify(service).createZipFile(exportData);
		}

		@Test
		void shouldWriteZipFileContentToOutputStream() {
			callService();

			verify(service).writeZipFileContent(zipFile, outputStream);
		}

		private void callService() {
			service.writeExport(VorgangHeaderTestFactory.ID, FILENAME_ID, outputStream);
		}
	}

	@Nested
	class TestCollectExportData {

		private MockedStatic<XdomeaNachrichtBuilder> xdomeaNachrichtBuilderMockedStatic;

		private static final String FILENAME_ID = UUID.randomUUID().toString();
		private static final String FILE_NAME = "file.zip";
		private final VorgangWithEingang vorgang = VorgangWithEingangTestFactory.create();
		private final VorgangType vorgangType = VorgangTypeTestFactory.create();
		private final NkAbgabeType kopfType = NkAbgabeTypeTestFactory.create();
		private final AbgabeAbgabe0401 abgabe = AbgabeAbgabe0401TestFactory.create();
		private final AkteType akteType = AkteTypeTestFactory.create();
		private final List<OzgFile> representations = List.of(OzgFileTestFactory.create());
		private final List<OzgFile> attachments = List.of(OzgFileTestFactory.create());
		private final List<DokumentType> representationsDokumentTypes = List.of(DokumentTypeTestFactory.create());
		private final List<DokumentType> attachmentsDokumentTypes = List.of(DokumentTypeTestFactory.create());
		private final List<HistorienProtokollInformationType> historienProtokollInformationTypes = List.of(
				HistorienProtokollInformationTypeTestFactory.create());
		private final KommentarsExportData kommentarsExportData = KommentarsExportDataTestFactory.create();
		private final PostfachMailExportData postfachMailExportData = PostfachMailExportDataTestFactory.create();

		@Mock
		private XdomeaNachrichtBuilder xdomeaNachrichtBuilder;

		@BeforeEach
		void setUp() {
			setUpVorgangService();
			setUpXdomeaNachrichtBuilder();
			setUpExportFileService();
			setUpExportHistorieService();
			setUpExportKommentarService();
			setUpExportNachrichtService();

			doReturn(FILE_NAME).when(service).buildXmlFilename(FILENAME_ID);
			doReturn(EingangHeaderTestFactory.FORM_ENGINE_NAME).when(service).getFormEngineName(vorgang);

		}

		private void setUpVorgangService() {
			when(exportVorgangService.getVorgang(VorgangHeaderTestFactory.ID)).thenReturn(vorgang);
			when(exportVorgangService.createVorgangType(vorgang)).thenReturn(vorgangType);
			when(exportVorgangService.createKopf(vorgang)).thenReturn(kopfType);
			when(exportVorgangService.createAkteType(vorgang)).thenReturn(akteType);
		}

		private void setUpXdomeaNachrichtBuilder() {
			xdomeaNachrichtBuilderMockedStatic = mockStatic(XdomeaNachrichtBuilder.class);
			when(xdomeaNachrichtBuilder.withKopf(kopfType)).thenReturn(xdomeaNachrichtBuilder);
			when(xdomeaNachrichtBuilder.withVorgang(vorgangType)).thenReturn(xdomeaNachrichtBuilder);
			when(xdomeaNachrichtBuilder.withAktenzeichen(akteType)).thenReturn(xdomeaNachrichtBuilder);
			when(xdomeaNachrichtBuilder.withRepresentations(representationsDokumentTypes)).thenReturn(xdomeaNachrichtBuilder);
			when(xdomeaNachrichtBuilder.withAttachments(attachmentsDokumentTypes)).thenReturn(xdomeaNachrichtBuilder);
			when(xdomeaNachrichtBuilder.withHistorie(historienProtokollInformationTypes)).thenReturn(xdomeaNachrichtBuilder);
			when(xdomeaNachrichtBuilder.withKommentare(kommentarsExportData.getDokumentTypes())).thenReturn(xdomeaNachrichtBuilder);
			when(xdomeaNachrichtBuilder.withPostfachMails(postfachMailExportData.getDokumentTypes())).thenReturn(xdomeaNachrichtBuilder);
			xdomeaNachrichtBuilderMockedStatic.when(XdomeaNachrichtBuilder::builder).thenReturn(xdomeaNachrichtBuilder);
			when(xdomeaNachrichtBuilder.build()).thenReturn(abgabe);
		}

		private void setUpExportFileService() {
			mockStreamToList(representations, stream -> when(exportFileService.getRepresentations(vorgang)).thenReturn(stream));
			mockStreamToList(attachments, stream -> when(exportFileService.getAttachments(vorgang)).thenReturn(stream));
			mockStreamToList(representationsDokumentTypes,
					stream -> when(exportFileService.createDokumentTypes(representations, EingangHeaderTestFactory.FORM_ENGINE_NAME))
							.thenReturn(stream));
			mockStreamToList(attachmentsDokumentTypes,
					stream -> when(exportFileService.createDokumentTypes(attachments, EingangHeaderTestFactory.FORM_ENGINE_NAME)).thenReturn(stream));
		}

		private void setUpExportHistorieService() {
			mockStreamToList(historienProtokollInformationTypes,
					stream -> when(exportHistorieService.createHistorienProtokollInformationTypes(vorgang)).thenReturn(stream));
		}

		private void setUpExportKommentarService() {
			when(exportKommentarService.createExportData(vorgang)).thenReturn(kommentarsExportData);
		}

		private void setUpExportNachrichtService() {
			when(exportNachrichtService.createExportData(vorgang)).thenReturn(postfachMailExportData);
		}

		@AfterEach
		void tearDown() {
			xdomeaNachrichtBuilderMockedStatic.close();
		}

		@Test
		void shouldLoadVorgang() {
			callService();

			verify(exportVorgangService).getVorgang(VorgangHeaderTestFactory.ID);
		}

		@Test
		void shouldGenerateXmlFilename() {
			callService();

			verify(service).buildXmlFilename(FILENAME_ID);
		}

		@Test
		void shouldGetRepresentations() {
			callService();

			verify(exportFileService).getRepresentations(vorgang);
		}

		@Test
		void shouldGetAttachments() {
			callService();

			verify(exportFileService).getAttachments(vorgang);
		}

		@Test
		void shouldCreateKopf() {
			callService();

			verify(exportVorgangService).createKopf(vorgang);
		}

		@Test
		void shouldCreateVorgangType() {
			callService();

			verify(exportVorgangService).createVorgangType(vorgang);
		}

		@Test
		void shouldCreateRepresentationsDokumentTypes() {
			callService();

			verify(exportFileService).createDokumentTypes(representations, EingangHeaderTestFactory.FORM_ENGINE_NAME);
		}

		@Test
		void shouldCreateAttachmentsDokumentTypes() {
			callService();

			verify(exportFileService).createDokumentTypes(attachments, EingangHeaderTestFactory.FORM_ENGINE_NAME);
		}

		@Test
		void shouldCreatePostfachMailsExportData() {
			callService();

			verify(exportNachrichtService).createExportData(vorgang);
		}

		@Test
		void shouldCreateAkteType() {
			callService();

			verify(exportVorgangService).createAkteType(vorgang);
		}

		@Test
		void shouldSetKopf() {
			callService();

			verify(xdomeaNachrichtBuilder).withKopf(kopfType);
		}

		@Test
		void shouldSetVorgang() {
			callService();

			verify(xdomeaNachrichtBuilder).withVorgang(vorgangType);
		}

		@Test
		void shouldSetRepresentations() {
			callService();

			verify(xdomeaNachrichtBuilder).withRepresentations(representationsDokumentTypes);
		}

		@Test
		void shouldSetAttachments() {
			callService();

			verify(xdomeaNachrichtBuilder).withAttachments(attachmentsDokumentTypes);
		}

		@Test
		void shouldSetPostfachMails() {
			callService();

			verify(xdomeaNachrichtBuilder).withPostfachMails(postfachMailExportData.getDokumentTypes());
		}

		@Test
		void shouldSetAktenzeichen() {
			callService();

			verify(xdomeaNachrichtBuilder).withAktenzeichen(akteType);
		}

		@Test
		void shouldSetHistorie() {
			callService();

			verify(xdomeaNachrichtBuilder).withHistorie(historienProtokollInformationTypes);
		}

		@Test
		void shouldCreateAbgabe() {
			callService();

			verify(xdomeaNachrichtBuilder).build();
		}

		@Test
		void shouldContainAbgabe() {
			var exportData = callService();

			assertThat(exportData.getAbgabe()).isEqualTo(abgabe);
		}

		@Test
		void shouldContainExportFilename() {
			var exportData = callService();

			assertThat(exportData.getExportFilename()).isEqualTo(FILE_NAME);
		}

		@Test
		void shouldContainExportExportFiles() {
			var exportData = callService();

			assertThat(exportData.getExportFiles())
					.hasSize(4)
					.containsAll(representations)
					.containsAll(attachments)
					.containsAll(kommentarsExportData.getAttachments())
					.containsAll(postfachMailExportData.getAttachments());
		}

		@Test
		void shouldExportKommentare() {
			callService();

			verify(exportKommentarService).createExportData(vorgang);
		}

		private ExportData callService() {
			return service.collectExportData(VorgangHeaderTestFactory.ID, FILENAME_ID);
		}
	}

	@DisplayName("Build xml Filename")
	@Nested
	class TestBuildXmlFilename {

		@Test
		void shouldMatchPattern() {
			var filename = service.buildXmlFilename(UUID.randomUUID().toString());

			assertThat(filename).matches(TestUtils.uuidRegexWithSuffix(ExportService.EXPORT_FILENAME_SUFFIX));
		}
	}

	@DisplayName("Create xml content")
	@Nested
	class TestCreateXmlContent {

		private static final String XML_STRING = "<xml>";

		@BeforeEach
		void init() {
			when(xdomeaXmlMarshaller.marshal(any())).thenReturn(XML_STRING);
		}

		@Test
		void shouldCallXmlMarshaller() {
			callService();

			verify(xdomeaXmlMarshaller).marshal(ExportDataTestFactory.ABGABE);
		}

		@Test
		void shouldReturnMarshalledString() {
			var result = callService();

			assertThat(result).isEqualTo(XML_STRING);
		}

		private String callService() {
			return service.createXmlContent(ExportDataTestFactory.ABGABE);
		}
	}

	@Nested
	class TestCreateZipFile {

		private static final String XML_FILE_CONTENT = "<xml></xml>";
		private static final String XML_STRING = "<xml>";
		private final ExportData exportData = ExportDataTestFactory.create();

		@Captor
		private ArgumentCaptor<ZipOutputStream> zipOutputStreamArgumentCaptor;

		@SneakyThrows
		@BeforeEach
		void setUp() {
			when(service.createXmlContent(ExportDataTestFactory.ABGABE)).thenReturn(XML_FILE_CONTENT);
			doNothing().when(service).putOzgFileIntoZip(any(OzgFile.class), any(ZipOutputStream.class));
		}

		@Test
		void shouldCreateXmlStringContent() {
			callService();

			verify(service).createXmlContent(ExportDataTestFactory.ABGABE);
		}

		@Test
		void shouldCreateZipEntry() throws IOException {
			callService();

			verify(service).putZipEntry(eq(ExportDataTestFactory.EXPORT_FILENAME), eq(XML_FILE_CONTENT),
					any(ZipOutputStream.class));
		}

		@Test
		void shouldCreateZipOutputStream() throws IOException {
			callService();

			verify(service).putZipEntry(eq(ExportDataTestFactory.EXPORT_FILENAME), eq(XML_FILE_CONTENT),
					zipOutputStreamArgumentCaptor.capture());
			assertThat(zipOutputStreamArgumentCaptor.getValue()).isInstanceOf(ZipOutputStream.class);
		}

		@Test
		void shouldWriteBytes() {
			var file = callService();

			assertThat(file).isNotEmpty().content().hasSizeGreaterThan(100);
		}

		@SneakyThrows
		@Test
		void shouldWritePdfFiles() {
			callService();

			verify(service).putOzgFileIntoZip(eq(ExportDataTestFactory.OZG_FILE), any(ZipOutputStream.class));
		}

		@Test
		void shouldThrowTechnicalException() throws IOException {
			doThrow(IOException.class).when(service).putZipEntry(anyString(), eq(XML_STRING), any(ZipOutputStream.class));

			assertThatThrownBy(this::callService).isInstanceOf(TechnicalException.class);
		}

		private File callService() {
			return service.createZipFile(exportData);
		}

	}

	@Nested
	class TestPutZipEntry {

		@Mock
		private ZipOutputStream zipOutputStream;

		@Captor
		private ArgumentCaptor<ZipEntry> zipEntryArgumentCaptor;

		@Captor
		private ArgumentCaptor<byte[]> fileDataArgumentCaptor;

		private final String FILE_NAME = "file.xml";
		private final String FILE_DATA = "xml content";

		@SneakyThrows
		@Test
		void shouldPutNextEntry() {
			callService();

			verify(zipOutputStream).putNextEntry(zipEntryArgumentCaptor.capture());
			assertThat(zipEntryArgumentCaptor.getValue().getName()).isEqualTo(FILE_NAME);
		}

		@SneakyThrows
		@Test
		void shouldWriteDataBytes() {
			callService();

			verify(zipOutputStream).write(fileDataArgumentCaptor.capture());
			assertThat(fileDataArgumentCaptor.getValue()).isEqualTo(FILE_DATA.getBytes(StandardCharsets.UTF_8));
		}

		@SneakyThrows
		@Test
		void shouldCloseEntry() {
			callService();

			verify(zipOutputStream).closeEntry();
		}

		@SneakyThrows
		private void callService() {
			service.putZipEntry(FILE_NAME, FILE_DATA, zipOutputStream);
		}
	}

	@Nested
	class TestPutOzgFileIntoZip {

		private MockedStatic<ExportFilenameGenerator> exportFilenameGenerator;

		@Mock
		private ZipOutputStream zipOutputStream;

		@Captor
		private ArgumentCaptor<ZipEntry> zipEntryArgumentCaptor;

		private final OzgFile ozgFile = OzgFileTestFactory.create();

		@BeforeEach
		void setUp() {
			exportFilenameGenerator = mockStatic(ExportFilenameGenerator.class);
			exportFilenameGenerator.when(() -> ExportFilenameGenerator.generateExportFilename(ozgFile)).thenReturn(OzgFileTestFactory.NAME);
		}

		@AfterEach
		void cleanUp() {
			exportFilenameGenerator.close();
		}

		@Test
		void shouldGenerateExportFilename() {
			callService();

			exportFilenameGenerator.verify(() -> ExportFilenameGenerator.generateExportFilename(ozgFile));
		}

		@SneakyThrows
		@Test
		void shouldPutNextEntry() {
			callService();

			verify(zipOutputStream).putNextEntry(zipEntryArgumentCaptor.capture());
			assertThat(zipEntryArgumentCaptor.getValue().getName()).isEqualTo(OzgFileTestFactory.NAME);
		}

		@Test
		void shouldCallExportFileService() {
			callService();

			verify(exportFileService).writeOzgFile(OzgFileTestFactory.ID, zipOutputStream);
		}

		@SneakyThrows
		@Test
		void shouldCloseEntry() {
			callService();

			verify(zipOutputStream).closeEntry();
		}

		@SneakyThrows
		private void callService() {
			service.putOzgFileIntoZip(ozgFile, zipOutputStream);
		}

	}

	@Nested
	class TestGetFormEngineName {

		@Test
		void shouldReturnEmptyStringOnNullEingang() {
			var vorgang = VorgangWithEingangTestFactory.createBuilder().eingang(null).build();

			var formEngineName = service.getFormEngineName(vorgang);

			assertThat(formEngineName).isEmpty();
		}

		@Test
		void shouldReturnEmptyStringOnNullEingangHeader() {
			var vorgang = VorgangWithEingangTestFactory.createBuilder().eingang(EingangTestFactory.createBuilder().header(null).build()).build();

			var formEngineName = service.getFormEngineName(vorgang);

			assertThat(formEngineName).isEmpty();
		}

		@Test
		void shouldReturnFormEngineName() {
			VorgangWithEingang vorgang = VorgangWithEingangTestFactory.create();

			var formEngineName = service.getFormEngineName(vorgang);

			assertThat(formEngineName).isEqualTo(EingangHeaderTestFactory.FORM_ENGINE_NAME);
		}
	}
}